;	PATHFINDING DEMO: This demo uses the AStar (A*) algorithm  
;	to find the shortest path between two points.
;	===============================================================
;	By Patrick Lester (4/12/02)

;	This demo was heavily influenced by the Path Search Demo written by
;	Bryan Stout. Info can be found at the following (registration
;	required) web pages.
;	article = http://www.gamasutra.com/features/19970801/pathfinding.htm
;	demo = http://www.gamasutra.com/features/19970801/oct96.zip

;	The basic A* method is to begin at the starting location and search
;	adjacent squares with the lowest F costs (see below) until you find
;	the target. The path is then generated by working backwards from the
;	target to the starting location, moving from each square to its
;	adjacent "parent" square.

;		F = G + H
;		G = distance from starting location to a given square by 
;			going along the path generated to get to there.
;		H = distance from the given square to the target square.

;	The best way to follow how this works is to do a step-by-step
;	search in this program (see below).

;	Instructions
;	------------
;	Demo finds the shortest path between the green square and the 
;	red square.
;	- Draw blue walls by left clicking on the screen. 
;	- Erase blue walls by left clicking on them.
;	- Move the red target around by right clicking on the screen.
;	- Move the green start square around by holding the "g" button
;		(g for green) down while left clicking on the screen.
;	- Start the pathfinding search by pressing the enter button.
;	- After the path is found, reset by clicking the mouse. You can 
;		draw new walls and then press enter to search again.
;	- Press 1 to do a step-by-step search. Continue pressing 1 for each 
;		new step. Press enter to finish a step-by-step search all at once.


;	Results
;	-------
;	- Path is marked by red dots.
;	- Squares highlighted in green were added to the "open list" of
;		squares to be checked, but were never fully checked.
;	- Squares highlighted in blue were fully checked and added to the
;		"closed list".
;	- Grey pointers point to a square's parent.
;	- F, G, and H costs are printed in each cell. The F cost is
;		printed in the upper left corner, G in lower left corner, and
;		H in lower right corner.
;	- Time to do search, in milliseconds, is printed to debug log.

;	Please email comments to Patrick Lester at pwlester@policyalmanac.org

	Include "includes/astarLibrary - Demo 1 (do not use).bb"	

	Graphics 800,600
	SetBuffer BackBuffer()
	HidePointer
	Gosub loadGraphics ; load graphics used in demo

;	Set pathfinderID and starting locations
	pathfinderID = 1
	startX = 3 : startY = 6 	;Set seeker location
	targetx = 12 : targety = 6  ;Set initial target location. This can
								;be changed by right-clicking on the map.

	
.mainloop	
	While Not KeyHit(1) ;while escape key isn't pressed	

;	Process user mouse and keyboard input
	Gosub userInput		
	
;	Draw stuff on the screen
	renderScreen(startX,startY,targetX,targetY,stepBystep)	
	
	Wend
	End


.loadGraphics

; 	Load the mouse cursor
	Global cursor = LoadImage("graphics\red_pointer.bmp")
	MaskImage cursor,255,255,255  ;set image transparant color to white
	MoveMouse 400,225 
			
;	Load grids
	Global grid = LoadImage("graphics\grid.bmp")		
		MaskImage grid,255,255,255
	Global greengrid = LoadImage("graphics\greengrid.bmp")		
		MaskImage greengrid,255,255,255			
	Global bluegrid = LoadImage("graphics\bluegrid.bmp")		
		MaskImage bluegrid,255,255,255
	Global dottedPath = LoadImage("graphics\path.bmp")		
		MaskImage dottedPath,255,255,255
	
;	Load blocks
	Global greenBlock = LoadImage("graphics\start.bmp") ;start location
	Global redBlock = LoadImage("graphics\end.bmp")	 ;target location
	Global wallBlock = LoadImage("graphics\wall.bmp")   ;wall
	
;	Load parentArrows - these point from a square to its parent in 
;	the pathfinding search.
	Dim parentArrow(9)
	For z = 1 To 8
		parentArrow(z) = LoadImage("graphics\arrow"+z+".bmp")	
		MaskImage parentArrow(z),255,255,255
	Next
	
;	Load numberstrip
	Global numbers = LoadAnimImage("graphics\SmallNumberStrip.bmp",4,5,0,12)	
	Return	


.userInput

	If (Not MouseDown(1)) And (Not MouseDown(2)) Then drawing = 0
	If (Not MouseDown(1)) And (Not MouseDown(2)) Then erasing = 0
	mouseXCoordinate = MouseX()/50
	mouseYCoordinate = MouseY()/50	
	
;	Draw walls (blue squares)
	If MouseDown(1) And (Not KeyDown(34)) Then
	If walkability(mouseXCoordinate,mouseYCoordinate) = walkable And erasing = 0 Then
	 	walkability(mouseXCoordinate,mouseYCoordinate) = unwalkable
		drawing = 1
	End If

;	Erase walls	
	If walkability(mouseXCoordinate,mouseYCoordinate) = unwalkable And drawing = 0 Then
	 	walkability(mouseXCoordinate,mouseYCoordinate) = walkable
		erasing = 1
	End If	
	End If;	If Mousedown(1) Then
		
;	Move red target square around
	If MouseHit(2) Then
		targetX = mouseXCoordinate
		targetY = mouseYCoordinate
	End If

;	Move starting square around
	If KeyDown(34) ; if g key is down while left clicking
	If MouseDown(1) Then
		startX = mouseXCoordinate
		startY = mouseYCoordinate
	End If
	End If

;	Start A* pathfinding search if return/enter key is hit
	If path = notfinished ;if path not searched
	If KeyHit(28) Then 
		startTime = MilliSecs()	;Record the time the search is begun.	
		path=findPath(pathfinderID,mapWidth,mapHeight,startX,startY,targetX,targetY)	

		;Calculate time to do search. Print to debuglog.
		endTime = MilliSecs() - startTime
		DebugLog "Total time to do pathfinding = " + endTime + " milliseconds."
		If path = found Then readPath(pathfinderID)	
		listArray(startX,startY) = 0 ;don't highlight the start square (aesthetics)	
	End If	
	End If

;	If "1" key is hit, start A* pathfinding search step-by-step
	If path = notfinished ;if path not searched
	If KeyHit(2) Then 
		FlushKeys
		stepByStep = True
			path=findPath(pathfinderID,mapWidth,mapHeight,startX,startY,targetX,targetY,stepBystep)
			If path = found Then 
				readPath(pathfinderID)
				listArray(startX,startY) = 0 ;don't highlight the start square (aesthetics)	
			End If
;			If escape key was hit before finishing path then reset	
			If path = notfinished Then Dim listArray(mapWidth+1,mapHeight+1)
		stepBystep = False	
	End If
	End If
	
;	Reset after finishing A* search if pressing mouse button or "1" or enter key 
	If path <> notfinished ;if path is done
	If (MouseDown(1) Or MouseDown(2) Or KeyHit(2) Or KeyHit(28)) Then 
		Dim listArray(17,13) ;clear out list data, which clears screen.
		path = notfinished
	End If	
	End If	

	Return	

	
.renderscreen	
Function renderScreen(startX,startY,targetX,targetY,stepByStep)	
	Cls
	
;	Draw starting location	
	DrawBlock greenBlock,startX*50,startY*50		

;	Draw target location		
	DrawBlock redBlock,targetX*50,targetY*50
	
	For x = 0 To 15 ;draw each of 16 tiles across the screen
	For y = 0 To 11 ;draw each of 12 tiles down the screen
	
;		Draw blue walls
		If walkability(x,y) = unwalkable Then DrawBlock wallBlock,x*50,y*50
				
;		Draw cells on open list (green), closed list (blue) and
;		part of the found path, if any (red dots).
		If listArray(x,y) = 1 Then DrawImage greengrid,x*50,y*50
		If listArray(x,y) = 2 Then DrawImage bluegrid,x*50,y*50
		If listArray(x,y) = 3 Then DrawImage dottedPath,x*50,y*50

;		Draw arrows that point to each square's parent.
		If listArray(x,y) = 1 Or listArray(x,y) = 2 
			If parentX(x,y) = x And parentY(x,y) = y - 1 Then 
				DrawImage parentArrow(1),x*50,y*50
			Else If parentX(x,y) = x+1 And parentY(x,y) = y - 1 Then 
				DrawImage parentArrow(2),x*50,y*50	
			Else If parentX(x,y) = x+1 And parentY(x,y) = y Then 
				DrawImage parentArrow(3),x*50,y*50
			Else If parentX(x,y) = x+1 And parentY(x,y) = y + 1 Then 
				DrawImage parentArrow(4),x*50,y*50				
			Else If parentX(x,y) = x And parentY(x,y) = y + 1 Then 
				DrawImage parentArrow(5),x*50,y*50
			Else If parentX(x,y) = x-1 And parentY(x,y) = y + 1 Then 
				DrawImage parentArrow(6),x*50,y*50
			Else If parentX(x,y) = x-1 And parentY(x,y) = y Then 
				DrawImage parentArrow(7),x*50,y*50
			Else If parentX(x,y) = x-1 And parentY(x,y) = y - 1 Then 
				DrawImage parentArrow(8),x*50,y*50
			End If	
		End If ;If listArray(x,y) = 1 Or listArray(x,y) = 2 

;		Print F, G and H costs in cells
		If listArray(x,y) > 0 ;if an open, closed, or path cell.
			G = Gcost(x,y) 			
			H = estimateHcost(x,y,targetx,targety)		
			F = G + H
		
			;Print F in upper left corner of cell (F=G+H)
			If F >= 100 Then DrawImage numbers,x*50+2,y*50+4,F/100		
			DrawImage numbers,x*50+7,y*50+4,(F Mod 100)/10
			DrawImage numbers,x*50+12,y*50+4,F Mod 10				

			;Print G in bottom left corner of cell (G=distance from start)
			If G >= 100 Then DrawImage numbers,x*50+2,y*50+40,G/100		
			DrawImage numbers,x*50+7,y*50+40,(G Mod 100)/10
			DrawImage numbers,x*50+12,y*50+40,G Mod 10

			;Print H in bottom right corner of cell (H=distance to target)
			If H >= 100 Then DrawImage numbers,x*50+32,y*50+40,H/100		
			DrawImage numbers,x*50+37,y*50+40,(H Mod 100)/10
			DrawImage numbers,x*50+42,y*50+40,H Mod 10			
		End If ;If listArray(x,y) > 0	


;		Draw the grid overlay.
		DrawImage grid,x*50,y*50

	Next
	Next

;	Draw the mouse (unless in step-by-step mode)
	If stepByStep = False Then DrawImage cursor,MouseX(),MouseY()		

	Flip
	End Function
	
	
	
	
		