;	PATHFINDING DEMO #3 - Islands and Dead Ends. Run the earlier demos 
;	first to get a better sense of how A* pathfinding works. 

;	This demo shows how larger maps can slow down A* star pathfinding and
;	introduces map preprocessing as an optimization tool. In this demo
;	we increase the speed of A* by preprocessing the map to identify
;	"islands" and "dead ends," which are defined as follows:

;	  * Islands: Islands are areas that are not reachable from anywhere 
;	else on the map. This can be an actual island or an area that is 
;	otherwise walled off. A map can have several "islands".
;		The benefit of pre-identifying islands is that it prevents A* 
;	from conducting long, fruitless searches (of nearly the entire map) 
;	looking for a nonexistent path to an unreachable target location.

;	  * Dead Ends: Pre-identifying dead ends allows A* to skip such 
;	areas when looking for a path between two points (except when a
;	particular dead end contains either the starting point or the
;	destination).

;	===============================================================

;	Instructions
;	------------
;	First do an A* search as usual (see below) of a lengthy path on the
;	screen. Then press escape and check the debuglog to see how long it
;	took. 
;		Then run the program again and press "p" to preprocess the map.
; 	Press enter or right click to start pathfinding. Left or right click
;	on the screen to do a similarly long search as before and quit and 
;	check the debuglog. 
;		Your search should have taken less time.
;	
;	- Press "p" to preprocess the map. Press "p" again to toggle
;		preprocessing off.
;	- Press "i" to show islands, "d" to show dead ends. Nothing will
;		happen if the map has not been preprocessed (see above). Press
;		"i" or "d" again to toggle drawing of these items off.
;	- Press enter or right click to activate the smiley sprite, then
;		left or right click on any reachable spot on the map. The smiley 
;		will use AStar to find the best path to that spot and then
;		go there. You may continue to left or right click to find more
;		paths. Press enter again to deactivate the sprite and return		
;		to map drawing mode.
;	- Draw blue walls by left clicking on the screen. If map has been
;		preprocessed, the preprocessing will be updated as you add
;		and delete walls (toggle drawing of deadends or islands 
;		on while drawing walls to see it in action).
;	- Erase blue walls by left clicking on them. If map has been
;		preprocessed, it will be updated as described above.
;	- Press escape to exit the program.

;	Please email comments to Patrick Lester at pwlester@policyalmanac.org

	mapWidth = 80
	mapHeight = 60	
	numberPeople = 2 ;one smiley, one chaser
	Include "includes/aStarLibrary.bb"
	Include "includes/pathPreprocessor.bb"

	If Not GfxModeExists(800,600,16) RuntimeError "Sorry, this program won't work with your graphics card."	
	Graphics 800,600;,0,1
	SetBuffer BackBuffer()
	HidePointer
	
;	Load graphics and previously saved map, if any
	Gosub loadGraphics
	Gosub loadMapData
	
;	Preprocess map: This is where we would normally put this command,
;	after the map data has been loaded. In this program we are 
;	toggling it on and off manually, so the line is commented out.	
;		preProcessMap(mapWidth,mapHeight)
	
;	Set starting locations
	Dim xLoc(2) : Dim yLoc(2) ;dim location variables
	xLoc(1) = 125 : yLoc(1) = 325 ;initial smiley location
	xLoc(2) = 725 : yLoc(2) = 325 ;initial chaser location
	Dim xPath(3) : Dim yPath(3)	;dim path location variables
	Dim pathStatus(3) ;dim path status (found, nonexistent, etc.)

	
.mainloop	
	While Not KeyHit(1) ;while escape key isn't pressed	

;	Process user mouse and keyboard input
	Gosub userInput		

;	Move smiley
	If smileyActivated = True Then Gosub moveSmiley
	
;	Move chaser
	If smileyActivated = True Then Gosub moveChaser	
	
;	Draw stuff on the screen
	Gosub renderScreen	
	
	Wend
	Gosub saveMapData
	End


.loadGraphics

; 	Load the mouse cursor
	cursor = LoadImage("graphics\red_pointer.bmp")
	MaskImage cursor,255,255,255  ;set image transparant color to white
	MoveMouse 400,225 
			
;	Load blocks
	wallBlock = LoadImage("graphics\wall x 10.bmp")   ;wall
	grid = LoadImage("graphics\grid.bmp")		
	MaskImage grid,255,255,255
	
;	Load smiley sprite
	smiley = LoadImage("graphics\smiley x 10.bmp")
		MaskImage smiley,0,0,0  ;set image transparant color to black
		MidHandle smiley	

;	Load chaser sprite
	chaser = LoadImage("graphics\ghost x 10.bmp")
		MaskImage chaser,0,0,0  ;set image transparant color to black
		MidHandle chaser

;	Create banks to store color information for dead end areas and islands
	areaColorRed = CreateBank(1)
	areaColorBlue = CreateBank(1)
	areaColorGreen = CreateBank(1)
	islandColorRed = CreateBank(1)
	islandColorBlue = CreateBank(1)
	islandColorGreen = CreateBank(1)	
		
	Return	


.AssignAreaColors ;This subroutine isn't necessary outside of this demo.
;It is used to color code our dead end areas.

;	Assign unique colors to each dead end area
	numberAreas = area(mapWidth,mapHeight)
	ResizeBank areaColorRed,numberAreas	+ 1
	ResizeBank areaColorBlue,numberAreas + 1	
	ResizeBank areaColorGreen,numberAreas + 1		
	SeedRnd MilliSecs()		
	For x = 1 To numberAreas
		PokeByte areaColorRed,x,Rand(200) + 25
		PokeByte areaColorBlue,x,Rand(200) + 25 
		PokeByte areaColorGreen,x,Rand(200) + 25 		
	Next		

	Return

	.assignNewAreaColors
	;Resize banks to allow for new areas
	newNumberAreas = area(mapWidth,mapHeight) 	
	ResizeBank areaColorRed,newNumberAreas + 1
	ResizeBank areaColorBlue,newNumberAreas + 1
	ResizeBank areaColorGreen,newNumberAreas + 1	

	;Set colors	for new areas, while retaining the same colors for 
	;the old areas.
	For x = numberAreas+1 To newNumberAreas
		PokeByte areaColorRed,x,Rand(200) + 25
		PokeByte areaColorBlue,x,Rand(200) + 25
		PokeByte areaColorGreen,x,Rand(200) + 25
	Next	
	numberAreas = area(mapWidth,mapHeight) 
	Return

	
.AssignIslandColors ;This subroutine isn't necessary outside of this demo.
;It is used to color code our islands.
;	If numberIslands <> island(mapWidth,mapHeight)

;	Resize banks to allow for new # of islands
	numberIslands = island(mapWidth,mapHeight)
	ResizeBank islandColorRed,numberIslands	+1
	ResizeBank islandColorBlue,numberIslands +1	
	ResizeBank islandColorGreen,numberIslands +1

;	The color of Island #1 defaults to black (0,0,0). For islands
;	beyond #1, assign a random color.			
	SeedRnd MilliSecs()	
	If numberIslands > 1	
	For x = 2 To numberIslands
		PokeByte islandColorRed,x,Rand(200) + 25
		PokeByte islandColorBlue,x,Rand(200) + 25 
		PokeByte islandColorGreen,x,Rand(200) + 25 		
	Next
	End If		

;	End If ;If numberIslands <> island(mapWidth,mapHeight)
	Return


.userInput

	If smileyActivated = False	
	
	
;1.	Draw walls (blue squares) while map not preprocessed (continuous draw)
	If MouseDown(1) And mapPreprocessed = False Then
	mouseXCoordinate = MouseX()/10
	mouseYCoordinate = MouseY()/10	
	If walkability(mouseXCoordinate,mouseYCoordinate) <> unwalkable And erasing = 0 Then
	 	walkability(mouseXCoordinate,mouseYCoordinate) = unwalkable
		drawing = 1
	End If


;2.	Erase walls	while map not preprocessed (continuous erase)
	If walkability(mouseXCoordinate,mouseYCoordinate) = unwalkable And drawing = 0 Then
	 	walkability(mouseXCoordinate,mouseYCoordinate) = walkable
		erasing = 1
	End If	
	
	FlushMouse
	End If;	If MouseDown(1) And mapPreprocessed = False

	;Stop continuous drawing/erasing when mouse button is up.	
	If (Not MouseDown(1)) And (Not MouseDown(2)) Then drawing = 0
	If (Not MouseDown(1)) And (Not MouseDown(2)) Then erasing = 0
	
	
;3.	Draw walls while map *is* preprocessed (one square at a time)
	;Note: Updates are only allowed one square at a time when map
	;has been preprocessed because we are updating the map on the
	;fly using the updateMap() function, which must be called after
	;each one square change.
	If MouseHit(1) And mapPreprocessed = True
	mouseXCoordinate = MouseX()/10
	mouseYCoordinate = MouseY()/10	

	If walkability(mouseXCoordinate,mouseYCoordinate) <> unwalkable Then
	 	walkability(mouseXCoordinate,mouseYCoordinate) = unwalkable


;4.	Erase walls	while map *is* preprocessed (one square at a time)
	Else If walkability(mouseXCoordinate,mouseYCoordinate) = unwalkable Then
	 	walkability(mouseXCoordinate,mouseYCoordinate) = walkable
	End If	

	;Since map has already been preprocessed, update it
	startTime = MilliSecs() ;Time the update  (not needed)

	;Call the updateMap() function.		
	updateMap(mouseXCoordinate,mouseYCoordinate,mapWidth,mapHeight)

	;Write time to do update to the debuglog (not needed)
	endTime = MilliSecs()
	DebugLog "Map updated. Time to do update = " + (endTime - startTime) + " milliseconds."	

	;Color code any new areas/islands (not needed)
	Gosub assignNewAreaColors
	Gosub AssignIslandColors
	
	End If ;MouseHit(1) And mapPreprocessed = True


;5.	Preprocess map by pressing "p"
	If KeyHit(25)

		;If mapPreprocessed = False then turn it on	
		If mapPreprocessed = False Then
			drawIslands = False
			drawDeadEnds = False	 
			preProcessMap(mapWidth,mapHeight) ;call the preProcessMap() function
			mapPreprocessed = True
			DebugLog "Map preprocessed"

		;If mapPreprocessed = True then toggle it off.
		Else If mapPreprocessed = True Then 
			mapPreprocessed = False
	
			;Erase dead ends	
			For x = 0 To mapWidth-1
			For y = 0 To mapHeight-1	
				If walkability(x,y) > 1 Then walkability(x,y) = walkable
			Next
			Next
		
			;Erase islands and existing areas
			Dim island(mapWidth,mapHeight)
			Dim area(mapWidth,mapHeight)

			DebugLog "Map preprocessing turned off."
		
		End If
						
	End If	


;6.	Show/draw preprocessed islands, if any, by pressing "i"
	If KeyHit(23) Then

		;If drawIslands = False, toggle it on.	
		If drawIslands = False Then
		
			drawIslands = True
			drawDeadEnds = False		
			Gosub AssignIslandColors ;color code the islands (not needed)

		;Otherwise, toggle it off.		
		Else
			drawIslands = False
		End If		
			
	End If	


;7.	Show/draw preprocessed dead ends, if any, by pressing "d"
	If KeyHit(32) Then

		;If drawDeadEnds = False, toggle it on.
		If drawDeadEnds = False Then 
			
			drawDeadEnds = True	;toggle drawing on
			drawIslands = False				
			Gosub AssignAreaColors ;color code the dead end areas (not needed)
			
		;If drawDeadEnds = True, toggle it off.
		Else
			drawDeadEnds = False ;toggle drawing off
		End If		

	End If	

	
;8.	Activate smiley sprite if return/enter or right mouse button is hit
	If KeyHit(28) Or MouseHit(2) Then 
		smileyActivated = True
		drawDeadEnds = False
		drawIslands = False		
		FlushKeys
		FlushMouse
		Gosub copyMapImage ;Copy the map image into one big image.
	End If

	Else ;If smileyActivated = True

	
;9.	Deactivate smiley sprite if return/enter key is hit again.
	If KeyHit(28) Then 
		smileyActivated = False

;		Reset path status. Normally this wouldn't be necessary  
;		because anything that needs to be reset is reset when the findPath
;		function is called. We do it "manually" here because it is
;		possible for the user to hit return, exiting pathfinding mode
;		and returning to wall drawing, when the smiley is halfway through 
;		a path. If we didn't reset it, when we hit return again to 
;		reactivate pathfinding, the smiley would continue to use the
;		old pathfinding info, plowing straight through any new walls you
;		may have created since it last checked the findPath function.
		pathStatus(1) = notstarted
		pathStatus(2) = notstarted		
		FlushKeys
	End If			
		
	End If ;If smileyActivated = False

	Return	

	
.renderscreen	

;	If smiley hasn't been activated then draw each of the map components
;	separately.
	If smileyActivated = False Then 
		Cls
	
		For x = 0 To 79 ;draw each of 80 tiles across the screen
		For y = 0 To 59 ;draw each of 60 tiles down the screen
	
;			Draw blue walls
			If walkability(x,y) = unwalkable Then DrawBlock wallBlock,x*10,y*10
		 
;			Draw deadend areas, if applicable
			If drawDeadEnds = True And walkability(x,y) = deadend Then 
				Color PeekByte (areaColorRed,area(x,y)),PeekByte (areaColorBlue,area(x,y)),PeekByte (areaColorGreen,area(x,y))
				Rect x*10,y*10,10,10,1
			End If

;			Draw islands, if applicable
			If drawIslands = True And walkability(x,y) <> unwalkable Then 
				Color PeekByte (islandColorRed,island(x,y)),PeekByte (islandColorBlue,island(x,y)),PeekByte (islandColorGreen,island(x,y))
				Rect x*10,y*10,10,10,1
			End If
		
		Next
		Next
	
	Else ;If smiley is activated then draw one big map image (faster)	
		DrawBlock map,0,0
	End If

;	Draw the smiley sprite
	DrawImage smiley,xLoc(1),yLoc(1)

;	Draw the chaser sprite
	DrawImage chaser,xLoc(2),yLoc(2)

;	Draw the mouse
	DrawImage cursor,MouseX(),MouseY()		

	Flip
	Return


.copyMapImage ;This subroutine copies the map image into one large image for
;	faster rendering. This isn't really necessary, and it doesn't effect
;	pathfinding at all. It just makes map drawing faster because drawing
;	one big image is a lot faster than repeatedly drawing its several 
;	thousand component images (there are potentially 4800 wall blocks 
;	on a map).

	FreeImage map ;Free the previous image, if any, thereby saving
	;valuable video memory.
	map = CreateImage(800,600) ;create a new map image.
	SetBuffer ImageBuffer(map)	

	For x = 0 To 79 ;draw each of 80 tiles across the screen
	For y = 0 To 59 ;draw each of 60 tiles down the screen

;		Draw blue walls
		If walkability(x,y) = unwalkable Then DrawBlock wallBlock,x*10,y*10
	
	Next
	Next	
	
	SetBuffer BackBuffer()
	Return



.loadMapData ;Load any pre-existing map when launching the program.
	filein = ReadFile("myTerrainData.dat")
	If filein Then
		For x = 0 To 79
		For y = 0 To 59
			walkability(x,y) = ReadByte(filein)
			If walkability(x,y) > 1 Then walkability(x,y) = 0
		Next
		Next
		CloseFile(filein)	
		End If
	Return


.saveMapData ;Save the map when exiting the program.
	fileout = WriteFile("myTerrainData.dat")	
	For x = 0 To 79
	For y = 0 To 59
		If walkability(x,y) <> 1 Then walkability(x,y) = 0
		WriteByte(fileout,walkability(x,y))			
	Next
	Next
	CloseFile(fileout)	
	Return

	
.moveSmiley  ;This subroutine moves the smiley around on the screen. In
;	this case the findPath function is accessed via mouse clicks.
	
	ID = 1 ;ID of smiley sprite

;1.	Find Path: If smiley is active, any left or right click 
;	on the map will find a path and make him go there.
	If MouseHit(1) Or MouseHit(2)

		;Start a timer to show how long the search takes (you obviously 
		;wouldn't do this outside of a demo).
		startTime = MilliSecs()

		;Call the findPath function. Note that we use screen (pixel  
		;based) coordinates, unlike in earlier versions of this demo.
		pathStatus(ID) = findPath(ID,mapWidth,mapHeight,10,xLoc(ID),yLoc(ID),MouseX(),MouseY())

		;Calculate time to do search. Print to debuglog (you obviously 
		;wouldn't do this outside of a demo).
		endTime = MilliSecs() - startTime
		DebugLog "Total time to do pathfinding = " + endTime + " milliseconds."

	End If
	
;2.	Move smiley.
	velocity = 5
	If pathStatus(ID) = found Then Gosub moveSprite			
	
	Return


.moveChaser ;This subroutine moves the chaser/ghost around on the screen.
;	In this case the findPath function is accessed automatically when
;	the chaser reaches the end of his current path. The path info
;	is also updated occasionally.

	ID = 2 ;ID of chaser
	targetID = 1 ;ID of target (the smiley)

;1.	Find Path: If smiley and chaser are not at the same location on the 
;	screen and no path is currently active, find a new path.
	If Not (  xLoc(ID) = xLoc(targetID) And yLoc(ID) = yLoc(targetID)  ) 

;	If no path has been generated, generate one. Update it when
;	the chaser reaches its fifth step on the current path.	
	If pathStatus(ID) = notstarted Or pathlocation(ID) = 5 

;		Generate a new path and read the x and y coordinates of the
;		first step. Enter coordinates of smiley sprite (xLoc(1)/
;		yLoc(1)) as the target.
		pathStatus(ID) = findPath(ID,mapWidth,mapHeight,10,xLoc(ID),yLoc(ID),xLoc(targetID),yLoc(targetID))
		
	End If
	End If ;If Not (xLoc(ID) = xLoc(ID) And yLoc(ID) = yLoc(ID)) 

;2.	Move chaser.
	velocity = 3
	If pathStatus(ID) = found Then Gosub moveSprite		

	Return



.moveSprite  ;This subroutine moves the sprites around on the screen.

;1.	Read path information
	readPath(ID,xLoc(ID),yLoc(ID),10,velocity)

;2.	Move sprite. xLoc/yLoc = current location of sprite. xPath and
;	yPath = coordinates of next step on the path that were/are
;	read using the readPath function.
	If xLoc(ID) > xPath(ID) Then xLoc(ID) = xLoc(ID) - velocity
	If xLoc(ID) < xPath(ID) Then xLoc(ID) = xLoc(ID) + velocity
	If yLoc(ID) > yPath(ID) Then yLoc(ID) = yLoc(ID) - velocity		
	If yLoc(ID) < yPath(ID) Then yLoc(ID) = yLoc(ID) + velocity
	
;3.	When sprite reaches the end location square	(end of its current
;	path) ...		
	If pathLocation(ID) = pathLength(ID) Then 
		
;		Center the chaser in the square (not really necessary, but 
;		it looks a little better for the chaser, which moves in 3 pixel
;		increments and thus isn't always centered when it reaches its
;		target).
		If Abs(xLoc(ID) - xPath(ID)) <= 2 Then xLoc(ID) = xPath(ID)
		If Abs(yLoc(ID) - yPath(ID)) <= 2 Then yLoc(ID) = yPath(ID)

	End If ;If end location square has been reached.
	Return	