;===================================================================
;PREPROCESS MAP: These functions preprocess the map, allowing faster 
;A* searching.
;====================================================================
;Updated 4/28/02


;PREPROCESS MAP: This function preprocesses the entire map. Call it at
;the beginning of the game or whenever you load a new level or map.
;This function can easily take 200 milliseconds to perform, so it is
;not something you want to do in the middle of a game. Instead, for
;mid-game updates when your map changes, use the updateMap() function 
;below. It is much faster, only updating the part of the map affected
;by the change.
Function preProcessMap(mapWidth,mapHeight)

	identifyIslands(mapWidth,mapHeight) ;identify islands (see below)
	identifyDeadEnds(mapWidth,mapHeight,1,mapWidth,1,mapHeight,1) ;identify dead ends (see below)

End Function


;UPDATE MAP: This function dynamically updates the map when a single wall
;block is added or removed. Parameters are as follows:

;- mapWidth,mapHeight = same as usual
;- x,y = x,y coordinates on the walkability array of the area that has
;	been changed. Remember that the walkability array ranges from (0,0)
;	in the upper left corner to (mapWidth-1,mapHeight-1) in the bottom
;	right corner.

;The function should be called after each one block change (i.e. if three
;squares change from walkable to unwalkable, this function should be called
;three times -- one for each square -- immediately after the change occurs.)
Function updateMap(x,y,mapWidth,mapHeight)

	updateIslands(x,y,mapWidth,mapHeight)			
	updateDeadEnds(x,y,mapWidth,mapHeight)

End Function


;====================================================================
;The following functions are called by the above two functions. You
;do not need to call them directly ....


;IDENTIFY ISLANDS: This function identifies "islands" on the map. 
;Islands are areas that are not reachable from anywhere not on/in
;the same area. These can be actual islands or simply areas walled 
;off from the rest of the map.
Function identifyIslands(mapWidth,mapHeight)

	areaID = 0 : Dim island(mapWidth+1,mapHeight+1) 
	For x = 0 To mapWidth-1
	For y = 0 To mapHeight-1

	If walkability(x,y) <> unwalkable And island(x,y) = 0 Then
	
;			Do a flood fill (just like the paint can in a graphics
;			program) to find out which deadends belong in the
;			same area.
			areaID =  areaID + 1	
			floodfill(areaID,mapWidth,mapHeight,x,y,walkable,deadend,1)
	
	End If			
	Next
	Next
	island(mapWidth,mapHeight) = areaID ; record total number of islands	
	
End Function


;IDENTIFY DEAD ENDS: This function identifies dead ends on the map.
;The A* function will ignore these areas when searching for a path
;(except deadends that contain either the starting or target location).
Function identifyDeadEnds(mapWidth,mapHeight,minX,maxX,minY,maxY,mode)


;1.	Create a temporary array a bit bigger than walkability array. This
;	allows proper processing of edge squares.
	Dim tempWalkability(mapWidth+2,mapHeight+2)
	For x = 1 To mapWidth
	For y = 1 To mapHeight		
		tempWalkability(x,y) = walkability(x-1,y-1)
	Next
	Next
	For y = 0 To mapHeight+1 ;convert edges of array to unwalkable
		tempWalkability(0,y) = unwalkable
		tempWalkability(mapWidth+1,y) = unwalkable		
	Next	 
	For x = 0 To mapWidth+1
		tempWalkability(x,0) = unwalkable
		tempWalkability(x,mapHeight+1) = unwalkable		
	Next


;2.	Dead End Identication: Optimization Pass # 1
	For x = minX To maxX
	For y = minY To maxY

	If tempWalkability(x,y) = walkable

;a.		"Hard" corners (hard = squares bordering wall/unwalkable areas)
		For a = -1 To +1 Step 2 ;flip horizontal
		For b = -1 To +1 Step 2 ;flip vertical	
		
			If tempWalkability(x-a,y) = unwalkable And tempWalkability(x,y+b) = unwalkable And tempWalkability(x+a,y-b) = walkable
				tempWalkability(x,y) = deadend : deadEndFound = True
			End If	

		Next
		Next

;b.		Three & four sided "hard" dead ends
		adjacentBlocks = 0
		If tempWalkability(x-1,y) = unwalkable Then adjacentBlocks = adjacentBlocks + 1
		If tempWalkability(x,y-1) = unwalkable Then adjacentBlocks = adjacentBlocks + 1
	 	If tempWalkability(x,y+1) = unwalkable Then adjacentBlocks = adjacentBlocks + 1
		If tempWalkability(x+1,y) = unwalkable Then adjacentBlocks = adjacentBlocks + 1

		;If blocked on three or four sides, then it's a dead end.
		If adjacentBlocks >= 3 Then 
			tempWalkability(x,y) = deadend 
			deadEndFound = True
		End If

			
	End If			
	Next
	Next


;3.	Dead End Identication: Optimization Pass # 2
	Repeat
	deadEndFound = False
	For x = minX To maxX ;check each coordinate on the map
	For y = minY To maxY

	If tempWalkability(x,y) = walkable

;a.		Find three & four sided "soft" dead ends  ("soft" = squares 
;		bordering deadends or walls/unwalkable)
		adjacentBlocks = 0
	 	If tempWalkability(x-1,y-1) <> walkable Then adjacentBlocks = adjacentBlocks + 1
		If tempWalkability(x-1,y) <> walkable Then adjacentBlocks = adjacentBlocks + 1
		If tempWalkability(x-1,y+1) <> walkable Then adjacentBlocks = adjacentBlocks + 1
		If tempWalkability(x,y-1) <> walkable Then adjacentBlocks = adjacentBlocks + 1
	 	If tempWalkability(x,y+1) <> walkable Then adjacentBlocks = adjacentBlocks + 1
		If tempWalkability(x+1,y-1) <> walkable Then adjacentBlocks = adjacentBlocks + 1
		If tempWalkability(x+1,y) <> walkable Then adjacentBlocks = adjacentBlocks + 1
		If tempWalkability(x+1,y+1) <> walkable Then adjacentBlocks = adjacentBlocks + 1

		;If blocked on 7/8 sides, then it's a dead end.
		If adjacentBlocks >= 7 Then 
			tempWalkability(x,y) = deadend 
			deadEndFound = True
		End If

;b.		C shaped deadends (checking five squares)		
		For a = -1 To +1 Step 2 ;flip horizontal
		b = 1
			If tempWalkability(x-a,y-b) <> walkable And tempWalkability(x,y-b) <> walkable
			If tempWalkability(x-a,y) <> walkable And tempWalkability(x+a,y) = walkable
			If tempWalkability(x-a,y+b) <> walkable And tempWalkability(x,y+b) <> walkable 
				tempWalkability(x,y) = deadend : deadEndFound = True
			End If	
			End If
			End If		
		Next
		
;c.		U shaped deadends - same as above but rotated 90 degrees counterclockwise
		For b = -1 To +1 Step 2 ;flip horizontal
		a = 1
			If tempWalkability(x,y-b) = walkable
			If tempWalkability(x-a,y) <> walkable And tempWalkability(x+a,y) <> walkable
			If tempWalkability(x-a,y+b) <> walkable And tempWalkability(x,y+b) <> walkable And tempWalkability(x+a,y+b) <> walkable 
				tempWalkability(x,y) = deadend : deadEndFound = True
			End If	
			End If
			End If		
		Next
		
;d.		Process "long" corners
		For a = -1 To +1 Step 2 ;flip horizontal
		For b = -1 To +1 Step 2 ;flip vertical		

			;Long Corners soft
			If tempWalkability(x,y-b) = walkable And tempWalkability(x+a,y-b) = walkable
			If tempWalkability(x-a,y) <> walkable And tempWalkability(x+a,y) = walkable
			If tempWalkability(x-a,y+b) <> walkable And tempWalkability(x,y+b) <> walkable 
			If tempWalkability(x+a,y+b) <> walkable Or tempWalkability(x-a,y-b) <> walkable
				tempWalkability(x,y) = deadend : deadEndFound = True
			End If	
			End If
			End If
			End If

		Next
		Next
		
	End If ;If tempWalkability(x,y) = walkable			
	Next
	Next
	Until deadEndFound = False Or KeyDown(1)

		
;4.	Dead End Identication: Optimization Pass # 3	
	Repeat
	deadEndFound = False
	For x = minX To maxX ;check each coordinate on the map
	For y = minY To maxY
	If tempWalkability(x,y) = walkable
	For a = -1 To +1 Step 2 ;flip horizontal
	For b = -1 To +1 Step 2 ;flip vertical		

		;Thumbs 1
		If tempWalkability(x-a,y-b) = walkable And tempWalkability(x,y-b) = walkable And tempWalkability(x+a,y-b) <> unwalkable
		If tempWalkability(x-a,y) = unwalkable And tempWalkability(x+a,y) = walkable
		If tempWalkability(x-a,y+b) = unwalkable And tempWalkability(x,y+b) <> walkable
			tempWalkability(x,y) = deadend : deadEndFound = True
		End If
		End If
		End If
		
		;Thumbs 2 (same as above, but rotated clockwise 90 degrees)	
		If tempWalkability(x-a,y-b) = unwalkable And tempWalkability(x,y-b) = unwalkable And tempWalkability(x+a,y-b) = walkable
		If tempWalkability(x-a,y) <> walkable  And tempWalkability(x+a,y) = walkable
		If tempWalkability(x,y+b) = walkable And tempWalkability(x+a,y+b) <> unwalkable
			tempWalkability(x,y) = deadend : deadEndFound = True
		End If
		End If
		End If

		;Double Paths 1
		If tempWalkability(x-a,y-b) <> walkable And tempWalkability(x,y-b) <> walkable And tempWalkability(x+a,y-b) <> walkable
		If tempWalkability(x-a,y) <> walkable And tempWalkability(x+a,y) = walkable
		If tempWalkability(x,y+b) = walkable And tempWalkability(x+a,y+b) <> unwalkable
			tempWalkability(x,y) = deadend : deadEndFound = True
		End If
		End If
		End If
		
		;Double Paths 2 (same as above, but rotated clockwise 90 degrees)	
		If tempWalkability(x,y-b) <> walkable And tempWalkability(x+a,y-b) <> walkable
		If tempWalkability(x-a,y) = walkable And tempWalkability(x+a,y) <> walkable
		If tempWalkability(x-a,y+b) <> unwalkable And tempWalkability(x,y+b) = walkable And tempWalkability(x+a,y+b) <> walkable
			tempWalkability(x,y) = deadend : deadEndFound = True
		End If
		End If
		End If

	Next
	Next
	End If ;If tempWalkability(x,y) = walkable			
	Next
	Next
	Until deadEndFound = False Or KeyDown(1)		

		
;5.	Convert tempWalkability back to walkability
	For x = minX To maxX
	For y = minY To maxY
		walkability(x-1,y-1) = tempWalkability(x,y)
	Next
	Next

				
;6.	Assign Area IDs to each dead end area
	areaID = area(mapWidth,mapHeight)
	If mode = 1 Then Dim area(mapWidth+1,mapHeight+1) 
	For x = minX-1 To maxX-1 ;check each coordinate on the map
	For y = minY-1 To maxY-1
		If walkability(x,y) = deadend And area(x,y) = 0 Then
		
;			Do a flood fill (just like the paint can in a graphics
;			program) to find out which deadends belong in the
;			same area.
			areaID =  areaID + 1
			floodfill(areaID,mapWidth,mapHeight,x,y,deadend,deadend,2)
							
		End If
	Next
	Next
	area(mapWidth,mapHeight) = areaID ; record total number of areas

End Function


;==========================================================
;FLOOD FILL: This function is used to assign squares to a
;specific area. It is used by the identifyIslands function to 
;figure out which areas are part of the same "island". It
;is also used by the identifyDeadEnds function to figure
;out which deadends are part of the same deadend area. It works
;just like the paint can tool in a graphics program.
Function floodFill(areaID,mapWidth,mapHeight,startX,startY,seekFill1,seekFill2,floodFillID)

;1.	Reset some variables that need to be cleared
	Dim listArray(mapWidth,mapHeight)	

;2.	Add the starting location to the open list of squares to be checked.
	openListItems = 1
	openList(1) = 1
	openX(1) = startX : openY(1) = startY

;3.	Do the following until the open list is empty.
	Repeat

;	Pop the first item off the list
	parentXval = openX(openList(1)) : parentYVal = openY(openList(1)) ;record cell coordinates

;	Delete it from the open list and add to closed list
	openList(1) = openList(openListItems) ;put last item in slot #1	
	openListItems = openListItems - 1 ;reduce number of open list items by 1
	listArray(parentXval,parentYVal) = onClosedList ;add cell to closed list

;	Assign item to areaID
	If floodFillID = 1 Then island(parentXval,parentYVal) = areaID
	If floodFillID = 2 Then area(parentXval,parentYVal) = areaID

;	Check its adjacent cells.
	For b = parentYVal-1 To parentYVal+1
	For a = parentXval-1 To parentXval+1
		
;	If not off the map (do this first to avoid array out of bounds errors)
	If a <> -1 And b <> -1 And a <> mapWidth And b <> mapHeight
	
;	If not on the open or closed lists
	If listArray(a,b) = onNeitherList 	
	
;	If an orthogonal square of the right type(s)
	If (walkability(a,b) = seekFill1 Or walkability(a,b) = seekFill2) And (a=parentXVal Or b=parentYVal)	

;		Create a new open list item.
		squaresChecked = squaresChecked + 1
		m = openListItems+1 ;m = new item at end of heap					
		openList(m) = squaresChecked				
		openX(squaresChecked) = a : openY(squaresChecked) = b		
		openListItems = openListItems+1 ;add one to number of items on the open list

;		Change listArray to show that the new item is on the open list.
		listArray(a,b) = onOpenList

	End If ;If an orthogonal square of the right type(s)	
	End If ;If not already on the closed list	
	End If ;If not off the map.
	Next
	Next

;	Exit when open list is empty
	Until openListItems = 0
	End Function
	

;====================================
;UPDATE ISLANDS
Function updateIslands(x,y,mapWidth,mapHeight)


;1.	Merge Check: If an opening has been created, check to see if two
;	or more areas need to be merged.

;a.	Create a list of islands affected by the change. 
	Dim adjacentArea(10)		
	For a = x-1 To x+1 ; check adjacent squares
	For b = y-1 To y+1
	
;	If not off the map
	If a <> -1 And b <> -1 And a <> mapWidth And b <> mapHeight

;	If not an unwalkable square	
	If walkability(a,b) <> unwalkable
		
;	Go through list of adjacent areas ...
	For z = 1 To 9	
	If island(a,b) = adjacentArea(z) Then ;if already on list, exit
		Exit
	Else If adjacentArea(z) = 0 Then ;if not already on list, add it
		adjacentArea(z) = island(a,b)
		areasAffected = areasAffected + 1
		If adjacentArea(z) = 1 Then firstArea = True
		Exit	
	End If
	Next

	End If ;if not an unwalkable square
	End If ;if not off the map
	Next
	Next

;b.	If no more than one island is affected, then no merge is necessary.
	If walkability(x,y) = walkable
	If areasAffected < 2 Then 
		Return

;c.	Otherwise, merge the islands that are affected	
	Else
			
	For a = 0 To mapWidth-1
	For b = 0 To mapHeight-1
	For z = 1 To areasAffected

;	If square is in the effected area then merge them with the first
;	area on the list.	
	If island(a,b) = adjacentArea(z) Or (a=x And b=y) Then 
		If firstArea = True Then 
			island(a,b) = 1
		Else	
			island(a,b) = adjacentArea(1)	
		End If	
	End If	
	
	Next
	Next
	Next
	End If ;If areasAffected < 2 Then 

	End If ;If walkability(x,y) = walkable
	
	
;2.	Split Check: If a blockage has been created, check to see if one
;	area needs to be split into two....
	If walkability(x,y) = unwalkable Then
	
	For a = x-1 To x+1
	For b = y-1 To y+1
	
;	If not off the map
	If a <> -1 And b <> -1 And a <> mapWidth And b <> mapHeight

;	Don't consider the square itself
	If Not (a = x And b = y)

;	Add up the number of adjacent squares (usually 8 if not an edge square)
	totalBlocks = totalBlocks+1
	
		If walkability(a,b) <> unwalkable Then openBlocks = openBlocks+1

	End If
	End If	
	Next
	Next
	
	;If number of adjacent open blocks is 0-1 or 7-8 don't change anything.
	;We haven't created a split.
	If openBlocks < 2 Or openBlocks >= totalBlocks-1 Then Return
	

;b.	Otherwise, a full update is needed ...
	For a = 0 To mapWidth-1
	For b = 0 To mapHeight-1
	For z = 1 To areasAffected

;	If square is in the affected area then zero it out	
	If island(a,b) = adjacentArea(z) Then island(a,b) = 0	
	
	Next
	Next
	Next

;	Redesignate new islands	
	areaID = island(mapWidth,mapHeight)
	For x = 0 To mapWidth-1
	For y = 0 To mapHeight-1

	If walkability(x,y) <> unwalkable And island(x,y) = 0 Then
	
;			Do a flood fill (just like the paint can in a graphics
;			program) to find out which deadends belong in the
;			same area.
			If areaID = island(mapWidth,mapHeight) And firstArea = True Then 
				areaID = 1
				firstArea = False
			Else	
				areaID =  areaID + 1	
			End If	
			floodfill(areaID,mapWidth,mapHeight,x,y,walkable,deadend,1)
			If areaID = 1 Then areaID = island(mapWidth,mapHeight)
	
	End If			
	Next
	Next
	island(mapWidth,mapHeight) = areaID ; record total number of islands

	End If;If walkability(x,y) = unwalkable Then

End Function
	
;================================================================
;UPDATE DEAD ENDS
Function updateDeadEnds(x,y,mapWidth,mapHeight)	
	
;1.	Quick Add: If adding a wall, there is no reason to convert any
;	dead ends to walkable if we have not blocked a passage between two
;	adjacent walkable squares.
	If walkability(x,y) = unwalkable ;if adding a wall at the location
	For a = x-1 To x+1
	For b = y-1 To y+1
	
;	If not off the map
	If a <> -1 And b <> -1 And a <> mapWidth And b <> mapHeight

;	Don't consider the square itself
	If Not (a = x And b = y)

;	Add up the number of adjacent squares (usually 8 if not an edge square)
	totalBlocks = totalBlocks+1
	
		If walkability(a,b) = walkable Then openBlocks = openBlocks+1

	End If
	End If	
	Next
	Next
	
	;If number of adjacent blocks is 0-1 or 8 don't change anything. We
	;haven't created a blockage between two adjacent walkable squares.
	If openBlocks < 2 Or openBlocks = totalBlocks Then Return
	End If


;2.	Quick Remove: If removing a wall, check to see if a passage has
;	been opened between two or more adjacent dead end areas.
	If walkability(x,y) = walkable ;if removing a wall at the location
	For a = x-1 To x+1
	For b = y-1 To y+1
	
;	If not off the map
	If a <> -1 And b <> -1 And a <> mapWidth And b <> mapHeight

;	Don't consider the square itself
	If Not (a = x And b = y)
	
		If walkability(a,b) = walkable Then openBlocks = openBlocks+1
		If walkability(a,b) = deadend Then 
			If thisArea = 0 Then 
				deadEndAreas = 1
				thisArea = area(a,b)
			Else If thisArea <> area(a,b) Then ;2+ areas are adjacent
				deadEndAreas = 2
			End If
		End If	

	End If
	End If	
	Next
	Next
	
	;If number of adjacent openBlocks is < 2 then placing a deadend in
	;the location  will not block any passage. If there is 1 adjacent
	;deadend area then we can associate this deadend with that area.
	;However, if there is more than 1 adjacent deadend area then we
	;may have opened a passage between them, in which a full reprocess
	;of the affected areas is needed.	
	If openBlocks < 2 ;no passage blocked
	If deadEndAreas = 1 ;1 adjacent deadend area	
		walkability(x,y) = deadend 
		area(x,y) = thisArea 	
		Return
	End If ;If deadEndAreas = 1
	End If ;If openBlocks < 2		
	End If ;If walkability(x,y) = walkable 	


;3.	Full Update
	
;a.	Create a list of dead end areas affected by the change. 
	Dim adjacentArea(26)		
	For a = x-2 To x+2 ; check each square within 2 squares
	For b = y-2 To y+2
	
;	If not off the map
	If a <> -1 And b <> -1 And a <> mapWidth And b <> mapHeight

;	If not an unwalkable square	
	If walkability(a,b) <> unwalkable
		
;	Go through list of adjacent areas ...
	For z = 1 To 25	
	If area(a,b) = adjacentArea(z) Then ;if already on list, exit
		Exit
	Else If adjacentArea(z) = 0 Then ;if not already on list, add it
		adjacentArea(z) = area(a,b)
		areasAffected = areasAffected + 1
		Exit	
	End If
	Next

	End If ;if not an unwalkable square
	End If ;if not off the map
	Next
	Next
	If areasAffected = 0 Then Return

	
;b.	Figure out the range of the map that the affected areas fall in and
;	zero out all the dead ends in the affected area(s).
	minimumX = mapWidth : minimumY = mapHeight
	maximumX = 0 : maximumY = 0
	For a = 0 To mapWidth-1
	For b = 0 To mapHeight-1
	For z = 1 To areasAffected

;	If square is in the effected area ...	
	If area(a,b) = adjacentArea(z)

;	Measure range of area affected. This limits our update to the
;	specific part of the map that has been effected.	
	If a < minimumX Then minimumX = a
	If a > maximumX Then maximumX = a
	If b < minimumY Then minimumY = b
	If b > maximumY Then maximumY = b
	
;	Zero out deadends
	If walkability(a,b) = deadend Then 
		walkability(a,b) = walkable
		area(a,b) = 0	
	End If	
	
	End If ;If area(a,b) = adjacentArea(z)	
	
	Next
	Next
	Next	


;c.	Update dead ends: Check everything within 1 of range of areas affected
	identifyDeadEnds(mapWidth,mapHeight,minimumX,maximumX+2,minimumY,maximumY+2,2)

	End Function
